/*
 * Copyright © 2010 Matthew Gregan <kinetik@flim.org>
 *
 * This program is made available under the ISC license.  See the
 * accompanying file LICENSE for details.
 */
#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "halloc.h"
#include "nestegg/nestegg.h"

/* EBML Elements */
#define ID_EBML                 0x1a45dfa3
#define ID_EBML_VERSION         0x4286
#define ID_EBML_READ_VERSION    0x42f7
#define ID_EBML_MAX_ID_LENGTH   0x42f2
#define ID_EBML_MAX_SIZE_LENGTH 0x42f3
#define ID_DOCTYPE              0x4282
#define ID_DOCTYPE_VERSION      0x4287
#define ID_DOCTYPE_READ_VERSION 0x4285

/* Global Elements */
#define ID_VOID                 0xec
#define ID_CRC32                0xbf

/* WebMedia Elements */
#define ID_SEGMENT              0x18538067

/* Seek Head Elements */
#define ID_SEEK_HEAD            0x114d9b74
#define ID_SEEK                 0x4dbb
#define ID_SEEK_ID              0x53ab
#define ID_SEEK_POSITION        0x53ac

/* Info Elements */
#define ID_INFO                 0x1549a966
#define ID_TIMECODE_SCALE       0x2ad7b1
#define ID_DURATION             0x4489

/* Cluster Elements */
#define ID_CLUSTER              0x1f43b675
#define ID_TIMECODE             0xe7
#define ID_BLOCK_GROUP          0xa0
#define ID_SIMPLE_BLOCK         0xa3

/* BlockGroup Elements */
#define ID_BLOCK                0xa1
#define ID_BLOCK_DURATION       0x9b
#define ID_REFERENCE_BLOCK      0xfb

/* Tracks Elements */
#define ID_TRACKS               0x1654ae6b
#define ID_TRACK_ENTRY          0xae
#define ID_TRACK_NUMBER         0xd7
#define ID_TRACK_UID            0x73c5
#define ID_TRACK_TYPE           0x83
#define ID_FLAG_ENABLED         0xb9
#define ID_FLAG_DEFAULT         0x88
#define ID_FLAG_LACING          0x9c
#define ID_TRACK_TIMECODE_SCALE 0x23314f
#define ID_LANGUAGE             0x22b59c
#define ID_CODEC_ID             0x86
#define ID_CODEC_PRIVATE        0x63a2

/* Video Elements */
#define ID_VIDEO                0xe0
#define ID_PIXEL_WIDTH          0xb0
#define ID_PIXEL_HEIGHT         0xba
#define ID_PIXEL_CROP_BOTTOM    0x54aa
#define ID_PIXEL_CROP_TOP       0x54bb
#define ID_PIXEL_CROP_LEFT      0x54cc
#define ID_PIXEL_CROP_RIGHT     0x54dd
#define ID_DISPLAY_WIDTH        0x54b0
#define ID_DISPLAY_HEIGHT       0x54ba

/* Audio Elements */
#define ID_AUDIO                0xe1
#define ID_SAMPLING_FREQUENCY   0xb5
#define ID_CHANNELS             0x9f
#define ID_BIT_DEPTH            0x6264

/* Cues Elements */
#define ID_CUES                 0x1c53bb6b
#define ID_CUE_POINT            0xbb
#define ID_CUE_TIME             0xb3
#define ID_CUE_TRACK_POSITIONS  0xb7
#define ID_CUE_TRACK            0xf7
#define ID_CUE_CLUSTER_POSITION 0xf1
#define ID_CUE_BLOCK_NUMBER     0x5378

/* EBML Types */
#define TYPE_MASTER             0
#define TYPE_UINT               1
#define TYPE_FLOAT              2
#define TYPE_STRING             3
#define TYPE_BINARY             4
#define TYPE_INT                5

#define LIMIT_STRING            (1 << 20)
#define LIMIT_BINARY            (1 << 24)
#define LIMIT_BLOCK             (1 << 30)
#define LIMIT_FRAME             (1 << 28)

/* Field Flags */
#define DESC_FLAG_NONE          0
#define DESC_FLAG_MULTI         (1 << 0)
#define DESC_FLAG_SUSPEND       (1 << 1)
#define DESC_FLAG_OFFSET        (1 << 2)

/* Lacing Constants */
#define LACING_NONE             0
#define LACING_XIPH             1
#define LACING_FIXED            2
#define LACING_EBML             3

struct ebml_binary {
  unsigned char * data;
  size_t length;
};

struct ebml_list_node {
  struct ebml_list_node * next;
  void * data;
};

struct ebml_list {
  struct ebml_list_node * head;
  struct ebml_list_node * tail;
};

struct ebml_vint {
  uint64_t value;
  size_t length;
};

struct ebml_svint {
  int64_t value;
  size_t length;
};

struct ebml_element {
  struct ebml_vint id;
  struct ebml_vint size;
};

struct list_node {
  struct list_node * previous;
  struct ebml_element_desc * node;
  unsigned char * data;
};

/* Element Descriptor */
struct ebml_element_desc {
  char const * name;
  uint64_t id;
  unsigned int type;
  size_t offset;
  unsigned int flags;
  struct ebml_element_desc * elements;
  size_t size;
  size_t data_offset;
};

/* EBML Definitions */
struct ebml {
  uint64_t ebml_version;
  uint64_t ebml_read_version;
  uint64_t ebml_max_id_length;
  uint64_t ebml_max_size_length;
  char * doctype;
  uint64_t doctype_version;
  uint64_t doctype_read_version;
};

/* Matroksa Definitions */
struct seek {
  struct ebml_binary id;
  uint64_t position;
};

struct seek_head {
  struct ebml_list seek;
};

struct info {
  uint64_t timecode_scale;
  double duration;
};

struct block_group {
  uint64_t duration;
  int64_t reference_block;
};

struct cluster {
  uint64_t timecode;
  struct ebml_list block_group;
};

struct video {
  uint64_t pixel_width;
  uint64_t pixel_height;
  uint64_t pixel_crop_bottom;
  uint64_t pixel_crop_top;
  uint64_t pixel_crop_left;
  uint64_t pixel_crop_right;
  uint64_t display_width;
  uint64_t display_height;
};

struct audio {
  double sampling_frequency;
  uint64_t channels;
  uint64_t bit_depth;
};

struct track_entry {
  uint64_t number;
  uint64_t uid;
  uint64_t type;
  uint64_t flag_enabled;
  uint64_t flag_default;
  uint64_t flag_lacing;
  double timecode_scale;
  char * language;
  char * codec_id;
  struct ebml_binary codec_private;
  struct video video;
  struct audio audio;
};

struct tracks {
  struct ebml_list track_entry;
};

struct cue_track_positions {
  uint64_t track;
  uint64_t cluster_position;
  uint64_t block_number;
};

struct cue_point {
  uint64_t time;
  struct ebml_list cue_track_positions;
};

struct cues {
  struct ebml_list cue_point;
};

struct segment {
  struct ebml_list seek_head;
  struct info info;
  struct ebml_list cluster;
  struct tracks tracks;
  struct cues cues;
};

struct pool_ctx {
  char dummy;
};

struct saved_state {
  int64_t stream_offset;
  struct list_node * ancestor;
  struct ebml_element element;
};

struct nestegg {
  nestegg_io io;
  nestegg_log log;
  struct pool_ctx * alloc_pool;
  struct ebml_element element;
  struct list_node * ancestor;
  struct ebml ebml;
  struct segment segment;
  int64_t segment_offset;
  unsigned int track_count;
};

struct frame {
  unsigned char * data;
  size_t length;
  struct frame * next;
};

struct nestegg_packet {
  uint64_t track;
  uint64_t timecode;
  struct frame * frame;
};

#define E_FIELD(ID, TYPE, STRUCT, FIELD) \
  { #ID, ID, TYPE, offsetof(STRUCT, FIELD), DESC_FLAG_NONE, NULL, 0, 0 }
#define E_MASTER(ID, TYPE, STRUCT, FIELD) \
  { #ID, ID, TYPE, offsetof(STRUCT, FIELD), DESC_FLAG_MULTI, FIELD ## _elements, \
      sizeof(struct FIELD), 0 }
#define E_SINGLE_MASTER_O(ID, TYPE, STRUCT, FIELD) \
  { #ID, ID, TYPE, offsetof(STRUCT, FIELD), DESC_FLAG_OFFSET, FIELD ## _elements, 0, \
      offsetof(STRUCT, FIELD ## _offset) }
#define E_SINGLE_MASTER(ID, TYPE, STRUCT, FIELD) \
  { #ID, ID, TYPE, offsetof(STRUCT, FIELD), DESC_FLAG_NONE, FIELD ## _elements, 0, 0 }
#define E_SUSPEND(ID, TYPE) \
  { #ID, ID, TYPE, 0, DESC_FLAG_SUSPEND, NULL, 0, 0 }
#define E_LAST \
  { NULL, 0, 0, 0, DESC_FLAG_NONE, NULL, 0, 0 }

/* EBML Element Lists */
static struct ebml_element_desc ebml_elements[] = {
  E_FIELD(ID_EBML_VERSION, TYPE_UINT, struct ebml, ebml_version),
  E_FIELD(ID_EBML_READ_VERSION, TYPE_UINT, struct ebml, ebml_read_version),
  E_FIELD(ID_EBML_MAX_ID_LENGTH, TYPE_UINT, struct ebml, ebml_max_id_length),
  E_FIELD(ID_EBML_MAX_SIZE_LENGTH, TYPE_UINT, struct ebml, ebml_max_size_length),
  E_FIELD(ID_DOCTYPE, TYPE_STRING, struct ebml, doctype),
  E_FIELD(ID_DOCTYPE_VERSION, TYPE_UINT, struct ebml, doctype_version),
  E_FIELD(ID_DOCTYPE_READ_VERSION, TYPE_UINT, struct ebml, doctype_read_version),
  E_LAST
};

/* WebMedia Element Lists */
static struct ebml_element_desc seek_elements[] = {
  E_FIELD(ID_SEEK_ID, TYPE_BINARY, struct seek, id),
  E_FIELD(ID_SEEK_POSITION, TYPE_UINT, struct seek, position),
  E_LAST
};

static struct ebml_element_desc seek_head_elements[] = {
  E_MASTER(ID_SEEK, TYPE_MASTER, struct seek_head, seek),
  E_LAST
};

static struct ebml_element_desc info_elements[] = {
  E_FIELD(ID_TIMECODE_SCALE, TYPE_UINT, struct info, timecode_scale),
  E_FIELD(ID_DURATION, TYPE_FLOAT, struct info, duration),
  E_LAST
};

static struct ebml_element_desc block_group_elements[] = {
  E_SUSPEND(ID_BLOCK, TYPE_BINARY),
  E_FIELD(ID_BLOCK_DURATION, TYPE_UINT, struct block_group, duration),
  E_FIELD(ID_REFERENCE_BLOCK, TYPE_INT, struct block_group, reference_block),
  E_LAST
};

static struct ebml_element_desc cluster_elements[] = {
  E_FIELD(ID_TIMECODE, TYPE_UINT, struct cluster, timecode),
  E_MASTER(ID_BLOCK_GROUP, TYPE_MASTER, struct cluster, block_group),
  E_SUSPEND(ID_SIMPLE_BLOCK, TYPE_BINARY),
  E_LAST
};

static struct ebml_element_desc video_elements[] = {
  E_FIELD(ID_PIXEL_WIDTH, TYPE_UINT, struct video, pixel_width),
  E_FIELD(ID_PIXEL_HEIGHT, TYPE_UINT, struct video, pixel_height),
  E_FIELD(ID_PIXEL_CROP_BOTTOM, TYPE_UINT, struct video, pixel_crop_bottom),
  E_FIELD(ID_PIXEL_CROP_TOP, TYPE_UINT, struct video, pixel_crop_top),
  E_FIELD(ID_PIXEL_CROP_LEFT, TYPE_UINT, struct video, pixel_crop_left),
  E_FIELD(ID_PIXEL_CROP_RIGHT, TYPE_UINT, struct video, pixel_crop_right),
  E_FIELD(ID_DISPLAY_WIDTH, TYPE_UINT, struct video, display_width),
  E_FIELD(ID_DISPLAY_HEIGHT, TYPE_UINT, struct video, display_height),
  E_LAST
};

static struct ebml_element_desc audio_elements[] = {
  E_FIELD(ID_SAMPLING_FREQUENCY, TYPE_FLOAT, struct audio, sampling_frequency),
  E_FIELD(ID_CHANNELS, TYPE_UINT, struct audio, channels),
  E_FIELD(ID_BIT_DEPTH, TYPE_UINT, struct audio, bit_depth),
  E_LAST
};

static struct ebml_element_desc track_entry_elements[] = {
  E_FIELD(ID_TRACK_NUMBER, TYPE_UINT, struct track_entry, number),
  E_FIELD(ID_TRACK_UID, TYPE_UINT, struct track_entry, uid),
  E_FIELD(ID_TRACK_TYPE, TYPE_UINT, struct track_entry, type),
  E_FIELD(ID_FLAG_ENABLED, TYPE_UINT, struct track_entry, flag_enabled),
  E_FIELD(ID_FLAG_DEFAULT, TYPE_UINT, struct track_entry, flag_default),
  E_FIELD(ID_FLAG_LACING, TYPE_UINT, struct track_entry, flag_lacing),
  E_FIELD(ID_TRACK_TIMECODE_SCALE, TYPE_FLOAT, struct track_entry, timecode_scale),
  E_FIELD(ID_LANGUAGE, TYPE_STRING, struct track_entry, language),
  E_FIELD(ID_CODEC_ID, TYPE_STRING, struct track_entry, codec_id),
  E_FIELD(ID_CODEC_PRIVATE, TYPE_BINARY, struct track_entry, codec_private),
  E_SINGLE_MASTER(ID_VIDEO, TYPE_MASTER, struct track_entry, video),
  E_SINGLE_MASTER(ID_AUDIO, TYPE_MASTER, struct track_entry, audio),
  E_LAST
};

static struct ebml_element_desc tracks_elements[] = {
  E_MASTER(ID_TRACK_ENTRY, TYPE_MASTER, struct tracks, track_entry),
  E_LAST
};

static struct ebml_element_desc cue_track_positions_elements[] = {
  E_FIELD(ID_CUE_TRACK, TYPE_UINT, struct cue_track_positions, track),
  E_FIELD(ID_CUE_CLUSTER_POSITION, TYPE_UINT, struct cue_track_positions, cluster_position),
  E_FIELD(ID_CUE_BLOCK_NUMBER, TYPE_UINT, struct cue_track_positions, block_number),
  E_LAST
};

static struct ebml_element_desc cue_point_elements[] = {
  E_FIELD(ID_CUE_TIME, TYPE_UINT, struct cue_point, time),
  E_MASTER(ID_CUE_TRACK_POSITIONS, TYPE_MASTER, struct cue_point, cue_track_positions),
  E_LAST
};

static struct ebml_element_desc cues_elements[] = {
  E_MASTER(ID_CUE_POINT, TYPE_MASTER, struct cues, cue_point),
  E_LAST
};

static struct ebml_element_desc segment_elements[] = {
  E_MASTER(ID_SEEK_HEAD, TYPE_MASTER, struct segment, seek_head),
  E_SINGLE_MASTER(ID_INFO, TYPE_MASTER, struct segment, info),
  E_MASTER(ID_CLUSTER, TYPE_MASTER, struct segment, cluster),
  E_SINGLE_MASTER(ID_TRACKS, TYPE_MASTER, struct segment, tracks),
  E_SINGLE_MASTER(ID_CUES, TYPE_MASTER, struct segment, cues),
  E_LAST
};

static struct ebml_element_desc top_level_elements[] = {
  E_SINGLE_MASTER(ID_EBML, TYPE_MASTER, nestegg, ebml),
  E_SINGLE_MASTER_O(ID_SEGMENT, TYPE_MASTER, nestegg, segment),
  E_LAST
};

#undef E_FIELD
#undef E_MASTER
#undef E_SINGLE_MASTER_O
#undef E_SINGLE_MASTER
#undef E_SUSPEND
#undef E_LAST

static struct pool_ctx *
pool_init(void)
{
  struct pool_ctx * pool;

  pool = h_malloc(sizeof(*pool));
  if (!pool)
    abort();
  return pool;
}

static void
pool_destroy(struct pool_ctx * pool)
{
  h_free(pool);
}

static void *
pool_alloc(size_t size, struct pool_ctx * pool)
{
  void * p;

  p = h_malloc(size);
  if (!p)
    abort();
  hattach(p, pool);
  memset(p, 0, size);
  return p;
}

#if 0
static void
pool_free(void * p)
{
  h_free(p);
}
#endif

static void *
alloc(size_t size)
{
  void * p;

  p = calloc(1, size);
  if (!p)
    abort();
  return p;
}

static int
bare_read_vint(nestegg * ctx, struct ebml_vint * vint, int mask_first_bit)
{
  int r;
  unsigned char b;
  size_t maxlen = 8;
  unsigned int count = 1, mask = 1 << 7;

  r = ctx->io.read(&b, 1, ctx->io.userdata);
  if (r != 1)
    return r;

  while (count < maxlen) {
    if ((b & mask) != 0)
      break;
    mask >>= 1;
    count += 1;
  }

  vint->length = count;
  vint->value = b;

  if (mask_first_bit) {
    vint->value = b & ~mask;
  }

  while (--count) {
    r = ctx->io.read(&b, 1, ctx->io.userdata);
    if (r != 1)
      return r;
    vint->value <<= 8;
    vint->value |= b;
  }

  return 1;
}

static int
read_id(nestegg * ctx, struct ebml_vint * vint)
{
  return bare_read_vint(ctx, vint, 0);
}

static int
read_vint(nestegg * ctx, struct ebml_vint * vint)
{
  return bare_read_vint(ctx, vint, 1);
}

static int
read_svint(nestegg * ctx, struct ebml_svint * svint)
{
  int r;
  struct ebml_vint vint;
  int64_t svint_subtr[] = {
    0x3f, 0x1fff,
    0xfffff, 0x7ffffff,
    0x3ffffffffLL, 0x1ffffffffffLL,
    0xffffffffffffLL, 0x7fffffffffffffLL
  };

  r = bare_read_vint(ctx, &vint, 1);
  if (r != 1)
    return r;
  svint->value = vint.value - svint_subtr[vint.length - 1];
  svint->length = vint.length;
  return r;
}

static int
read_uint(nestegg * ctx, uint64_t * val, uint64_t length)
{
  unsigned char b;
  int r;

  if (length == 0 || length > 8)
    return -1;
  r = ctx->io.read(&b, 1, ctx->io.userdata);
  if (r != 1)
    return r;
  *val = b;
  while (--length) {
    r = ctx->io.read(&b, 1, ctx->io.userdata);
    if (r != 1)
      return r;
    *val <<= 8;
    *val |= b;
  }
  return 1;
}

static int
read_int(nestegg * ctx, int64_t * val, uint64_t length)
{
  int r;
  uint64_t uval, base;

  r = read_uint(ctx, &uval, length);
  if (r != 1)
    return r;

  if (length < sizeof(int64_t)) {
    base = 1;
    base <<= length * 8 - 1;
    if (uval >= base) {
        base = 1;
        base <<= length * 8;
    } else {
      base = 0;
    }
    *val = uval - base;
  } else {
    *val = (int64_t) uval;
  }

  return 1;
}

static int
read_float(nestegg * ctx, double * val, uint64_t length)
{
  union {
    uint64_t u;
    float f;
    double d;
  } value;
  int r;

  /* length == 10 not implemented */
  if (length != 4 && length != 8)
    return -1;
  r = read_uint(ctx, &value.u, length);
  if (r != 1)
    return r;
  if (length == 4)
    *val = value.f;
  else
    *val = value.d;
  return 1;
}

static int
read_string(nestegg * ctx, char ** val, uint64_t length)
{
  char * str;
  int r;

  if (length == 0 || length > LIMIT_STRING)
    return -1;
  str = pool_alloc(length + 1, ctx->alloc_pool);
  r = ctx->io.read((unsigned char *) str, length, ctx->io.userdata);
  if (r != 1)
    return r;
  str[length] = '\0';
  *val = str;
  return 1;
}

static int
read_binary(nestegg * ctx, struct ebml_binary * val, uint64_t length)
{
  if (length == 0 || length > LIMIT_BINARY)
    return -1;
  val->data = pool_alloc(length, ctx->alloc_pool);
  val->length = length;
  return ctx->io.read(val->data, length, ctx->io.userdata);
}

static int
is_ancestor_element(struct ebml_element el, struct list_node * ancestor)
{
  struct ebml_element_desc * element;

  for (; ancestor; ancestor = ancestor->previous)
    for (element = ancestor->node; element->id; ++element)
      if (element->id == el.id.value)
        return 1;

  return 0;
}

static struct ebml_element_desc *
find_element(struct ebml_element el, struct ebml_element_desc * elements)
{
  struct ebml_element_desc * element;

  for (element = elements; element->id; ++element)
    if (element->id == el.id.value)
      return element;

  return NULL;
}

static void
ctx_push(nestegg * ctx, struct ebml_element_desc * ancestor, void * data)
{
  struct list_node * item;

  item = alloc(sizeof(*item));
  item->previous = ctx->ancestor;
  item->node = ancestor;
  item->data = data;
  ctx->ancestor = item;
}

static void
ctx_pop(nestegg * ctx)
{
  struct list_node * item;

  item = ctx->ancestor;
  ctx->ancestor = item->previous;
  free(item);
}

static int
ctx_save(nestegg * ctx, struct saved_state * s)
{
  s->stream_offset = ctx->io.tell(ctx->io.userdata);
  if (s->stream_offset < 0)
    return -1;
  s->ancestor = ctx->ancestor;
  s->element = ctx->element;
  return 0;
}

static int
ctx_restore(nestegg * ctx, struct saved_state * s)
{
  int r;

  r = ctx->io.seek(s->stream_offset, NESTEGG_SEEK_SET, ctx->io.userdata);
  if (r != 0)
    return -1;
  ctx->ancestor = s->ancestor;
  ctx->element = s->element;
  return 0;
}

static int
peek_element(nestegg * ctx, struct ebml_element * el)
{
  int r;

  if (ctx->element.id.value) {
    *el = ctx->element;
    return 1;
  }

  r = read_id(ctx, &ctx->element.id);
  if (r != 1)
    return r;

  r = read_vint(ctx, &ctx->element.size);
  if (r != 1)
    return r;

  *el = ctx->element;

  return 1;
}

static int
read_element(nestegg * ctx, struct ebml_element * el)
{
  int r;
  struct ebml_vint zero = { 0, 0 };

  r = peek_element(ctx, el);
  if (r != 1)
    return r;

  ctx->element.id = zero;
  ctx->element.size = zero;

  return 1;
}

static int
parse(nestegg * ctx)
{
  int r;
  int64_t * data_offset;
  struct ebml_element el;
  struct ebml_element_desc * element;
  struct ebml_list * list;
  struct ebml_list_node * node, * oldtail;
  void * p, * pp;

  /* loop until we need to return:
     - hit suspend point
     - parse complete
     - error occurred */

  /* loop over elements at current level reading them
     if sublevel found, push ctx onto stack and "continue"
     if sublevel ended, pop ctx off stack and "continue" */

  for (;;) {
    r = peek_element(ctx, &el);
    if (r != 1)
      break;

    element  = find_element(el, ctx->ancestor->node);
    if (element) {
      if (element->flags & DESC_FLAG_SUSPEND) {
        assert(element->type == TYPE_BINARY);
        ctx->log(ctx, NESTEGG_LOG_DEBUG, "suspend parse at %llx\n", el.id.value);
        r = 1;
        break;
      }

      r = read_element(ctx, &el);
      if (r != 1)
        break;

      if (element->flags & DESC_FLAG_OFFSET) {
        data_offset = (int64_t *) (ctx->ancestor->data + element->data_offset);
        *data_offset = ctx->io.tell(ctx->io.userdata);
        if (*data_offset < 0)
          return -1;
      }

      if (element->type == TYPE_MASTER) {
        if (element->flags & DESC_FLAG_MULTI) {

          ctx->log(ctx, NESTEGG_LOG_DEBUG, "multi master element %llx (%s)\n",
                   el.id.value, element->name);

          list = (struct ebml_list *) (ctx->ancestor->data + element->offset);

          node = pool_alloc(sizeof(*node), ctx->alloc_pool);
          node->data = pool_alloc(element->size, ctx->alloc_pool);

          oldtail = list->tail;
          if (oldtail)
            oldtail->next = node;
          list->tail = node;
          if (!list->head)
            list->head = node;

          ctx->log(ctx, NESTEGG_LOG_DEBUG, " -> using data %p\n", node->data);

          ctx_push(ctx, element->elements, node->data);
        } else {
          ctx->log(ctx, NESTEGG_LOG_DEBUG, "single master element %llx (%s)\n",
                   el.id.value, element->name);
          ctx->log(ctx, NESTEGG_LOG_DEBUG, " -> using data %p (%u)\n",
                   ctx->ancestor->data + element->offset, element->offset);
          p = (void *) (ctx->ancestor->data + element->offset);
          ctx_push(ctx, element->elements, p);
        }
        continue;
      } else {
        ctx->log(ctx, NESTEGG_LOG_DEBUG, "element %llx (%s) -> %p (%u)\n",
                 el.id.value, element->name,
             ctx->ancestor->data + element->offset, element->offset);
        pp = (void *) (ctx->ancestor->data + element->offset);
        switch (element->type) {
        case TYPE_UINT:
          r = read_uint(ctx, pp, el.size.value);
          break;
        case TYPE_FLOAT:
          r = read_float(ctx, pp, el.size.value);
          break;
        case TYPE_STRING:
          r = read_string(ctx, pp, el.size.value);
          break;
        case TYPE_BINARY:
          r = read_binary(ctx, pp, el.size.value);
          break;
        case TYPE_INT:
          r = read_int(ctx, pp, el.size.value);
          break;
        default:
          assert(0);
        }
        if (r != 1)
          return r;
      }
    } else if (is_ancestor_element(el, ctx->ancestor->previous)) {
      ctx->log(ctx, NESTEGG_LOG_DEBUG, "parent element %llx\n", el.id.value);
      ctx_pop(ctx);
    } else {
      r = read_element(ctx, &el);
      if (r != 1)
        break;

      if (el.id.value != ID_VOID && el.id.value != ID_CRC32)
        ctx->log(ctx, NESTEGG_LOG_DEBUG, "unknown element %llx\n", el.id.value);
      r = ctx->io.seek(el.size.value, NESTEGG_SEEK_CUR, ctx->io.userdata);
      if (r != 0)
        return -1;
    }
  }

  if (r != 1) {
    while(ctx->ancestor)
      ctx_pop(ctx);
  }

  return r;
}

static uint64_t
xiph_lace_value(unsigned char ** np)
{
  uint64_t lace;
  uint64_t value;
  unsigned char * p = *np;

  lace = *p++;
  value = lace;
  while (lace == 255) {
    lace = *p++;
    value += lace;
  }

  *np = p;

  return value;
}

static int
read_xiph_lace_value(nestegg * ctx, uint64_t * value, size_t * consumed)
{
  int r;
  uint64_t lace;

  r = read_uint(ctx, &lace, 1);
  if (r != 1)
    return r;
  *consumed += 1;

  *value = lace;
  while (lace == 255) {
    r = read_uint(ctx, &lace, 1);
    if (r != 1)
      return r;
    *consumed += 1;
    *value += lace;
  }

  return 1;
}

static int
read_xiph_lacing(nestegg * ctx, size_t block, size_t read, uint64_t n, uint64_t * sizes)
{
  int r;
  size_t i = 0;
  uint64_t sum = 0;

  while (--n) {
    r = read_xiph_lace_value(ctx, &sizes[i], &read);
    if (r != 1)
      return r;
    sum += sizes[i];
    i += 1;
  }

  if (read + sum > block)
    return -1;

  /* last frame is the remainder of the block */
  sizes[i] = block - read - sum;
  return 1;
}

static int
read_ebml_lacing(nestegg * ctx, size_t block, size_t read, uint64_t n, uint64_t * sizes)
{
  int r;
  struct ebml_svint slace;
  struct ebml_vint lace;
  uint64_t sum;
  size_t i = 0;

  r = read_vint(ctx, &lace);
  if (r != 1)
    return r;
  read += lace.length;

  sizes[i] = lace.value;
  sum = sizes[i];

  i += 1;
  n -= 1;

  while (--n) {
    r = read_svint(ctx, &slace);
    if (r != 1)
      return r;
    read += slace.length;
    sizes[i] = sizes[i - 1] + slace.value;
    sum += sizes[i];
    i += 1;
  }

  if (read + sum > block)
    return -1;

  /* last frame is the remainder of the block */
  sizes[i] = block - read - sum;
  return 1;
}

static int
read_block(nestegg * ctx, uint64_t block_size, nestegg_packet ** data, uint64_t type)
{
  int r;
  int64_t timecode;
  nestegg_packet * pkt;
  struct cluster * cluster;
  struct ebml_vint track;
  struct frame * f, * last;
  uint64_t frame_sizes[256], cluster_tc, flags, frames, pts_ns, tc_scale;
  unsigned int i, lacing;
  size_t consumed = 0;

  *data = NULL;

  if (block_size > LIMIT_BLOCK)
    return -1;

  tc_scale = ctx->segment.info.timecode_scale;
  if (tc_scale == 0)
    tc_scale = 1000000; /* XXX hack */

  cluster = ctx->segment.cluster.tail->data;
  cluster_tc = cluster->timecode;

  r = read_vint(ctx, &track);
  if (r != 1)
    return r;

  if (track.value == 0 || track.value > ctx->track_count)
    return -1;

  consumed += track.length;

  r = read_int(ctx, &timecode, 2);
  if (r != 1)
    return r;

  consumed += 2;

  r = read_uint(ctx, &flags, 1);
  if (r != 1)
    return r;

  consumed += 1;

  pts_ns = (cluster_tc + timecode) * tc_scale;

  frames = 0;

  /* flags are different between block and simpleblock, but lacing is
     encoded the same way */
  lacing = (flags & 0x6) >> 1;

  switch (lacing) {
  case LACING_NONE:
    frames = 1;
    break;
  case LACING_XIPH:
  case LACING_FIXED:
  case LACING_EBML:
    r = read_uint(ctx, &frames, 1);
    if (r != 1)
      return r;
    consumed += 1;
    frames += 1;
  }

  if (frames > 256)
    return -1;

  switch (lacing) {
  case LACING_NONE:
    frame_sizes[0] = block_size - consumed;
    break;
  case LACING_XIPH:
    if (frames == 1)
      return -1;
    r = read_xiph_lacing(ctx, block_size, consumed, frames, frame_sizes);
    if (r != 1)
      return r;
    break;
  case LACING_FIXED:
    if ((block_size - consumed) % frames)
      return -1;
    for (i = 0; i < frames; ++i) {
      frame_sizes[i] = (block_size - consumed) / frames;
    }
    break;
  case LACING_EBML:
    if (frames == 1)
      return -1;
    r = read_ebml_lacing(ctx, block_size, consumed, frames, frame_sizes);
    if (r != 1)
      return r;
    break;
  }

  pkt = alloc(sizeof(*pkt));
  pkt->track = track.value - 1;
  pkt->timecode = pts_ns;

  ctx->log(ctx, NESTEGG_LOG_DEBUG, "%sblock t %lld pts %f f %llx frames: ",
           type == ID_SIMPLE_BLOCK ? "simple" : "", pkt->track, pts_ns / 1e9, flags);

  /* XXX sanity check unlaced frame sizes against total block size */

  last = NULL;
  for (i = 0; i < frames; ++i) {
    if (frame_sizes[i] > LIMIT_FRAME) {
      nestegg_free_packet(pkt);
      return -1;
    }
    f = alloc(sizeof(*f));
    f->data = alloc(frame_sizes[i]);
    f->length = frame_sizes[i];
    r = ctx->io.read(f->data, frame_sizes[i], ctx->io.userdata);
    if (r != 1) {
      free(f->data);
      free(f);
      nestegg_free_packet(pkt);
      return -1;
    }

    ctx->log(ctx, NESTEGG_LOG_DEBUG, "%u ", frame_sizes[i]);

    if (!last) {
      pkt->frame = f;
    } else {
      last->next = f;
    }
    last = f;
  }
  ctx->log(ctx, NESTEGG_LOG_DEBUG, "\n");

  *data = pkt;

  return 1;
}

static struct track_entry *
find_track_entry(nestegg * ctx, unsigned int track)
{
  struct ebml_list_node * node;
  unsigned int tracks = 0;

  node = ctx->segment.tracks.track_entry.head;
  while (node) {
    if (track == tracks) {
      return node->data;
    }
    tracks += 1;
    node = node->next;
  }

  return NULL;
}

static uint64_t
buf_read_id(unsigned char const * p, size_t length)
{
  uint64_t id = 0;

  while (length--) {
    id <<= 8;
    id |= *p++;
  }

  return id;
}

static struct seek *
find_seek_for_id(struct ebml_list_node * seek_head, uint64_t id)
{
  struct ebml_list * head;
  struct ebml_list_node * seek;
  struct seek * s;

  while (seek_head) {
    head = seek_head->data;
    seek = head->head;

    while (seek) {
      s = seek->data;

      if (buf_read_id(s->id.data, s->id.length) == id) {
        return s;
      }

      seek = seek->next;
    }

    seek_head = seek_head->next;
  }

  return NULL;
}

static struct cue_point *
find_cue_point_for_tstamp(struct ebml_list_node * cue_point, uint64_t scale, uint64_t tstamp)
{
  struct cue_point *c, * prev = NULL;

  while (cue_point) {
    c = cue_point->data;

    if (!prev)
      prev = c;

    if (c->time * scale  > tstamp) {
      return prev;
    }

    prev = cue_point->data;
    cue_point = cue_point->next;
  }

  return NULL;
}

static int
is_suspend_element(uint64_t id)
{
  /* this could search the tree of elements for DESC_FLAG_SUSPEND */
  if (id == ID_SIMPLE_BLOCK || id == ID_BLOCK)
    return 1;
  return 0;
}

static void
null_log_callback(nestegg * ctx, unsigned int severity, char const * fmt, ...)
{
  if (ctx && severity && fmt)
    return;
}

int
nestegg_init(nestegg ** context, nestegg_io io, nestegg_log callback)
{
  int r;
  struct ebml_element el;
  struct ebml_list_node * track;
  nestegg * ctx = NULL;

  if (!(io.read && io.seek && io.tell))
    return -1;

  ctx = alloc(sizeof(*ctx));

  ctx->io = io;
  ctx->log = callback;
  ctx->alloc_pool = pool_init();

  if (!ctx->log)
    ctx->log = null_log_callback;

  r = peek_element(ctx, &el);
  if (r != 1) {
    nestegg_destroy(ctx);
    return -1;
  }

  if (el.id.value != ID_EBML) {
    nestegg_destroy(ctx);
    return -1;
  }

  ctx->log(ctx, NESTEGG_LOG_DEBUG, "ctx %p\n", ctx);

  ctx_push(ctx, top_level_elements, ctx);

  r = parse(ctx);

  if (r != 1) {
    nestegg_destroy(ctx);
    return -1;
  }

  /* XXX require: webm doctype, check ebml versions, check for segment info
     and tracks */
  if (!ctx->segment.tracks.track_entry.head) {
    nestegg_destroy(ctx);
    return -1;
  }


  track = ctx->segment.tracks.track_entry.head;
  ctx->track_count = 0;

  while (track) {
    ctx->track_count += 1;
    track = track->next;
  }

  *context = ctx;

  return 0;
}

void
nestegg_destroy(nestegg * ctx)
{
  while (ctx->ancestor)
    ctx_pop(ctx);
  pool_destroy(ctx->alloc_pool);
  free(ctx);
}

int
nestegg_duration(nestegg * ctx, uint64_t * duration)
{
  uint64_t tc_scale;

  tc_scale = ctx->segment.info.timecode_scale;
  if (tc_scale == 0)
    tc_scale = 1000000; /* XXX hack */

  *duration = (uint64_t) (ctx->segment.info.duration * tc_scale);
  return 0;
}

int
nestegg_track_count(nestegg * ctx, unsigned int * tracks)
{
  *tracks = ctx->track_count;
  return 0;
}

int
nestegg_track_seek(nestegg * ctx, unsigned int track, uint64_t tstamp)
{
  int r;
  struct cue_point * cue_point;
  struct cue_track_positions * pos;
  struct saved_state state;
  struct seek * found;
  uint64_t seek_pos;
  uint64_t tc_scale;
  struct ebml_list_node * node = ctx->segment.cues.cue_point.head;
  struct ebml_vint zero = { 0, 0 };

  /* If there are no cues loaded, check for cues element in the seek head
     and load it. */
  if (!node) {
    found = find_seek_for_id(ctx->segment.seek_head.head, ID_CUES);
    if (!found)
      return -1;

    /* Save old parser state. */
    r = ctx_save(ctx, &state);
    if (r != 0)
      return -1;

    /* Seek and set up parser state for segment-level element (Cues). */
    r = ctx->io.seek(ctx->segment_offset + found->position, NESTEGG_SEEK_SET, ctx->io.userdata);
    if (r != 0)
      return -1;
    ctx->element.id = zero;
    ctx->element.size = zero;

    ctx->ancestor = NULL;
    ctx_push(ctx, top_level_elements, ctx);
    ctx_push(ctx, segment_elements, &ctx->segment);
    /* XXX parser will run until block/simpleblock, EOF, or error. */
    r = parse(ctx);
    while (ctx->ancestor)
      ctx_pop(ctx);

    /* Reset parser state to original state and seek back to old position. */
    r = ctx_restore(ctx, &state);
    if (r != 0)
      return -1;
  }

  tc_scale = ctx->segment.info.timecode_scale;
  if (tc_scale == 0)
    tc_scale = 1000000; /* XXX hack */

  cue_point = find_cue_point_for_tstamp(ctx->segment.cues.cue_point.head, tc_scale, tstamp);
  if (!cue_point)
    return -1;

  node = cue_point->cue_track_positions.head;

  seek_pos = 0;

  while (node) {
    pos = node->data;
    if (pos->track - 1 == track) {
      seek_pos = pos->cluster_position;
      break;
    }
    node = node->next;
  }

  /* Seek and set up parser state for segment-level element (Cluster). */
  r = ctx->io.seek(ctx->segment_offset + seek_pos, NESTEGG_SEEK_SET, ctx->io.userdata);
  if (r != 0)
    return -1;
  ctx->element.id = zero;
  ctx->element.size = zero;

  while (ctx->ancestor)
    ctx_pop(ctx);

  ctx_push(ctx, top_level_elements, ctx);
  ctx_push(ctx, segment_elements, &ctx->segment);
  r = parse(ctx);
  if (r != 1)
    return -1;

  /* XXX sanity */
  if (!is_suspend_element(ctx->element.id.value))
      return -1;

  return 0;
}

int
nestegg_track_type(nestegg * ctx, unsigned int track)
{
  struct track_entry * entry;

  entry = find_track_entry(ctx, track);
  if (!entry)
    return -1;

  if (entry->type & 0x1)
    return NESTEGG_TRACK_VIDEO;

  if (entry->type & 0x2)
    return NESTEGG_TRACK_AUDIO;

  return -1;
}

struct bitmapinfoheader {
  int size;
  int width;
  int height;
  short planes;
  short bit_count;
  unsigned int compression;
  int size_image;
  int x_pels_per_meter;
  int y_pels_per_meter;
  int clr_used;
  int clr_important;
};

int
nestegg_track_codec_id(nestegg * ctx, unsigned int track)
{
  struct track_entry * entry;

  entry = find_track_entry(ctx, track);
  if (!entry)
    return -1;

  if (!entry->codec_id)
    return -1;

  if (strcmp(entry->codec_id, "V_VP8") == 0)
    return NESTEGG_CODEC_VP8;

  if (strcmp(entry->codec_id, "A_VORBIS") == 0)
    return NESTEGG_CODEC_VORBIS;

  /* XXX hack for old youtube encodes */
  if (strcmp(entry->codec_id, "V_MS/VFW/FOURCC") == 0 &&
      entry->codec_private.length >= 40) {
    struct bitmapinfoheader * bih = (struct bitmapinfoheader *) entry->codec_private.data;
    if (bih->compression == 0x30385056)
      return NESTEGG_CODEC_VP8;
  }

  return -1;
}

int
nestegg_track_codec_data_count(nestegg * ctx, unsigned int track,
                               unsigned int * count)
{
  struct track_entry * entry;
  unsigned char * p;

  *count = 0;

  entry = find_track_entry(ctx, track);
  if (!entry)
    return -1;

  if (!entry->codec_private.data)
    return -1;

  if (nestegg_track_codec_id(ctx, track) != NESTEGG_CODEC_VORBIS)
    return -1;

  p = entry->codec_private.data;
  *count = *p + 1;

  return 0;
}

int
nestegg_track_codec_data(nestegg * ctx, unsigned int track, unsigned int item,
                         unsigned char ** data, size_t * length)
{
  struct track_entry * entry;
  uint64_t sizes[3], total;
  unsigned char * p;
  unsigned int count, i;

  *data = NULL;
  *length = 0;

  entry = find_track_entry(ctx, track);
  if (!entry)
    return -1;

  if (nestegg_track_codec_id(ctx, track) != NESTEGG_CODEC_VORBIS)
    return -1;

  p = entry->codec_private.data;
  count = *p++ + 1;

  if (count > 3)
    return -1;

  i = 0;
  total = 0;
  while (--count) {
    sizes[i] = xiph_lace_value(&p);
    total += sizes[i];
    i += 1;
  }
  sizes[i] = entry->codec_private.length - total - (p - entry->codec_private.data);

  for (i = 0; i < item; ++i) {
    if (sizes[i] > LIMIT_FRAME)
      return -1;
    p += sizes[i];
  }
  *data = p;
  *length = sizes[item];

  return 0;
}

int
nestegg_track_video_params(nestegg * ctx, unsigned int track,
                           nestegg_video_params * params)
{
  struct track_entry * entry;

  memset(params, 0, sizeof(*params));

  entry = find_track_entry(ctx, track);
  if (!entry)
    return -1;

  if (nestegg_track_type(ctx, track) != NESTEGG_TRACK_VIDEO)
    return -1;

  params->width = entry->video.pixel_width;
  params->height = entry->video.pixel_height;

  params->crop_bottom = entry->video.pixel_crop_bottom;
  params->crop_top = entry->video.pixel_crop_top;
  params->crop_left = entry->video.pixel_crop_left;
  params->crop_right = entry->video.pixel_crop_right;

  params->display_width = entry->video.display_width;
  if (params->display_width == 0)
    params->display_width = entry->video.pixel_width;

  params->display_height = entry->video.display_height;
  if (params->display_height == 0)
    params->display_height = entry->video.pixel_height;

  return 0;
}

int
nestegg_track_audio_params(nestegg * ctx, unsigned int track,
                           nestegg_audio_params * params)
{
  struct track_entry * entry;

  memset(params, 0, sizeof(*params));

  entry = find_track_entry(ctx, track);
  if (!entry)
    return -1;

  if (nestegg_track_type(ctx, track) != NESTEGG_TRACK_AUDIO)
    return -1;

  params->rate = entry->audio.sampling_frequency;
  params->channels = entry->audio.channels;
  params->depth = entry->audio.bit_depth;

  return 0;
}

int
nestegg_read_packet(nestegg * ctx, nestegg_packet ** pkt)
{
  int r;
  struct ebml_element el;

  *pkt = NULL;

  for (;;) {
    r = peek_element(ctx, &el);
    if (r != 1) {
      return r;
    }

    /* any suspend fields must be handled here */
    if (is_suspend_element(el.id.value)) {
      r = read_element(ctx, &el);
      if (r != 1) {
        return r;
      }

      r = read_block(ctx, el.size.value, pkt, el.id.value);
      return r;
    }

    r =  parse(ctx);
    if (r != 1)
      return r;
  }

  return 1;
}

void
nestegg_free_packet(nestegg_packet * pkt)
{
  struct frame * frame;

  while (pkt->frame) {
    frame = pkt->frame;
    pkt->frame = frame->next;
    free(frame->data);
    free(frame);
  }

 free(pkt);
}

int
nestegg_packet_track(nestegg_packet * pkt, unsigned int * track)
{
  *track = pkt->track;
  return 0;
}

int
nestegg_packet_tstamp(nestegg_packet * pkt, uint64_t * tstamp)
{
  *tstamp = pkt->timecode;
  return 0;
}

int
nestegg_packet_count(nestegg_packet * pkt, unsigned int * count)
{
  struct frame * f = pkt->frame;

  *count = 0;

  while (f) {
    *count += 1;
    f = f->next;
  }

  return 0;
}

int
nestegg_packet_data(nestegg_packet * pkt, unsigned int item,
                    unsigned char ** data, size_t * length)
{
  struct frame * f = pkt->frame;
  unsigned int count = 0;

  *data = NULL;
  *length = 0;

  while (f) {
    if (count == item) {
      *data = f->data;
      *length = f->length;
      return 0;
    }
    count += 1;
    f = f->next;
  }

  return -1;
}
